function [M, A] = assemble_mass_and_stiffness(mesh)
    Vm = zeros(3, 3*size(mesh.F, 2));
    Va = zeros(size(Vm));
    I = zeros(size(Vm));
    J = zeros(size(Vm));

    for kk=1:size(mesh.F, 2)
        a = mesh.P(:, mesh.F(2, kk)) - mesh.P(:, mesh.F(1, kk));
        b = mesh.P(:, mesh.F(3, kk)) - mesh.P(:, mesh.F(2, kk));
        c = mesh.P(:, mesh.F(1, kk)) - mesh.P(:, mesh.F(3, kk));
        d = abs(det([a, c]));
        ab = a.'*b;
        bc = b.'*c;
        ca = c.'*a;
        
        Vm(:, 3*kk-2:3*kk) = d/24 * [2, 1, 1; ...
                                     1, 2, 1; ...
                                     1, 1, 2];
        Va(:, 3*kk-2:3*kk) = 0.5/d * [-bc-ab,     bc,     ab; ...
                                          bc, -ca-bc,     ca; ...
                                          ab,     ca, -ab-ca];
        
        I(:, 3*kk-2:3*kk) = [mesh.F(:, kk), mesh.F(:, kk), mesh.F(:, kk)];
        J(:, 3*kk-2:3*kk) = I(:, 3*kk-2:3*kk).';
    end
    
    M = sparse(I(:), J(:), Vm(:));
    A = sparse(I(:), J(:), Va(:));
end