function meshf = mesh_refine(mesh)
    P = mesh.P;
    B = mesh.B;
    F = mesh.F;

    [dp, np] = size(P);
    [df, nf] = size(F);

    D = zeros(1, np);
    for ii=1:nf
        for jj=1:df
            D(F(jj, ii)) = D(F(jj, ii)) + 1;
        end
    end
    D(B ~= 1) = D(B ~= 1) + 1;

    S = cumsum([0, D]);  % first shift into edge end point for start point
    npe = S(end)/2;
    E = zeros(1, 2*npe); % edge end point indices
    C = zeros(1, 2*npe); % centre point indices
    Pf = zeros(dp, np+npe); % extended point coordinates
    Pf(:, 1:np) = P;
    Bf = zeros(1, np+npe); % extended point boundary flags
    Bf(1, 1:np) = B;

    next = mod(1:df, df) + 1;
    cntr = np;

    for ii=1:nf
        for jj=1:df
            jjp = next(jj);
            a = F(jj, ii);
            b = F(jjp, ii);
            for kk=1+S(a):S(a+1)
                if E(kk) == b
                    Bf(C(kk)) = 1;
                    break;
                elseif E(kk) == 0
                    cntr = cntr + 1;
                    for ll=1:dp
                        Pf(ll, cntr) = .5*(Pf(ll, a) + Pf(ll, b));
                    end
                    Bf(cntr) = max(Bf(a), Bf(b));

                    E(kk) = b;
                    C(kk) = cntr;

                    for ll=1+S(b):S(b+1)
                        if E(ll) == 0
                            E(ll) = a;
                            C(ll) = cntr;
                            break;
                        end
                    end
                    break;
                end
            end
        end
    end

    meshf.P = Pf;
    meshf.B = Bf;

    Ff = zeros(df, 4*nf);
    m = zeros(df, 1);
    elem = 0;

    for ii=1:nf
        for jj=1:df
            jjp = next(jj);
            a = F(jj, ii);
            b = F(jjp, ii);

            for kk=1+S(a):S(a+1)
                if E(kk) == b
                    m(jj) = C(kk);
                    break;
                end
            end
        end

        elem = elem + 1;
        Ff(1, elem) = F(1, ii); Ff(2, elem) = m(1); Ff(3, elem) = m(3);
        elem = elem + 1;
        Ff(1, elem) = m(1); Ff(2, elem) = F(2, ii); Ff(3, elem) = m(2);
        elem = elem + 1;
        Ff(1, elem) = m(3); Ff(2, elem) = m(2); Ff(3, elem) = F(3, ii);
        elem = elem + 1;
        Ff(1, elem) = m(2); Ff(2, elem) = m(3); Ff(3, elem) = m(1);
    end

    meshf.F = Ff;
end
